function [trialsInBestRun, Bestrun, Worstg, Worstg1, n, example, S, bestDesignSummary, bestdesign, MaxW, Pow, Eff] = ...
    steep(example, omega, temperature, ...
          maxK, maxKK, Lambda, g1reps, maxL, num_runs, numStages)

% Example of call:
% [trialsInBestRun, Bestrun, Worstg, Worstg1, n, example, S, bestDesignSummary, bestdesign, MaxW, Pow, Eff] = ...
% Example 1:        steep(1, .5, .05, 3, 10, 5, 50, 1000, 100, 5);
%                   steep(1, 0, .05, 3, 10, 5, 50, 1000, 100, 5); (max power)
%                   steep(1, 1, .005, 3, 10, 5, 50, 1000, 10, 5); (max efficiency)
                                     
% Example 2:        steep(2, .5, .05, 3, 10, 5, 50, 1000, 100, 5);
%                   steep(2, 0, .05, 3, 10, 5, 50, 1000, 100, 5); (max power) 
%                   steep(2, 1, .005, 3, 10, 5, 50, 1000, 10, 5); (max efficiency)

% Example 3:        steep(3, .5, .05, 3, 10, 5, 50, 1000, 100, 5);
%                   steep(3, 0, .05, 3, 10, 5, 50, 1000, 100, 5); (max power) 
%                   steep(3, 1, .0015, 3, 10, 5, 50, 1000, 10, 5); (max efficiency)


% INPUTS:
% example - set to do the examples in the paper
% omega - choose in [0,1]; 0 means pure LOF problem, 1 means pure
    % estimation problem
% temperature = intitial temperature; set so that no more than one half of accepted states
    % are improvements, in the first stage.  If too many are accepted,
    % lower the temoperature.
% maxK = number of attempts to find a better g1, with lambda decreasing each time
% maxKK = number of descent steps to carry out
% Lambda controls the maximum step size in steepest descent
% g1reps = number of times to run steepest descent, from different starting
    % points each time
% max_L - this is 'L' in the paper
% num_runs = number of runs to be carried out
% numStages = minimum number of temperature changes

% Set the design space; compute mu0 and U
if(example == 1)
    link = 'l';
	n = 40;
    eta = 5;
    maxm = .4422124;
    % Set the design space:
    age = (30:75)'; 
    S = [[zeros(length(age),1),age];[ones(length(age),1),age]];
    T = S(:,1);
    age = S(:,2);
    mu = 2.3262 - 6.2646*T - .1051*age + .2057*T.*age;
    N = length(age);
    U = [ones(N,1), T, age];
    theta = regress(mu, U);
    q = length(theta);
    mu0 = U*theta;
elseif(example == 2)        
    link = 'l';
	n = 40;
    eta = 5;
    maxm = .7336636;
    % Set the design space:
    age = (30:75)'; 
    S = [[zeros(length(age),1),age];[ones(length(age),1),age]];
    T = S(:,1);
    age = S(:,2);
    mu = 2.3262 - 6.2646*T - .1051*age + .2057*T.*age;
    N = length(age);
    U = [ones(N,1), T, age, T.*age];
    theta = regress(mu, U); % recover mu    
    q = length(theta);
    mu0 = U*theta;
elseif(example == 3)        
    link = 'l';
    n = 20; 
    eta = 2;
    maxm = .0268293;
    % Set the design space:
    S = linspace(2,5,31)'; 
    N = length(S);
    alpha = 1;
    beta = 4;
	theta = [alpha; beta]; 
    q = length(theta);
    t = S;
    mu0 = alpha*log(t) - log(beta);
    U = [log(t) (-1/beta)*ones(N,1)];
end
ind = (1:N)';
% Do the annealing
runSummary = [];
bestDesignSummary = [];
MaxW = 0;
for run = 1:num_runs
    disp('Run'); disp(run);   
    % Run the annealing algorithm:
    [bestxi, worstg1, maxW, pow, eff, trialSummary, stageSummary] = ...
        annealrun(numStages, maxK, maxKK, Lambda, link, maxm, mu0, U, q, n, N, eta, g1reps, temperature, maxL, omega);
    %Store the best design from each run:
	bestDesignSummary = [bestDesignSummary, bestxi]; 
	%Summarize the current run:
	disp([blanks(6), 'run', blanks(6), 'MaxW', blanks(7), 'Pow', blanks(7), 'Eff']);
    runSummary = [runSummary; run, maxW, pow, eff];
    disp(runSummary);
    % Retain the output from the best run so far:
    if(maxW > MaxW) 
        MaxW = maxW;
        Pow = pow;
        Eff = eff;
        Worstg1 = worstg1;
        Bestxi = bestxi;
        trialsInBestRun = trialSummary;
        Bestrun = stageSummary;
    end
end


%% Finalize: 
% Summarize best design and worst g:
xi = Bestxi;
% Compute the power if the most efficient design was obtained:
if(omega == 1)  
indpos = ind(xi>0);
s = length(indpos);
outmat = zeros(s+1, g1reps);
    for count = 1:g1reps
        g1start = randn(s,1); % Random starting point;
        g1start = eta*g1start/norm(g1start);
        [g1, fval] = descent(g1start, maxK, maxKK, Lambda, link, mu0(indpos), xi(indpos), eta);
        outmat(:, count) = [g1; fval];
    end
        bestfval = min(outmat(s+1,:));
        Worstg1 = outmat((1:s),(outmat(s+1,:) == bestfval));
        Pow = normcdf(sqrt(2*n*bestfval) - norminv(.95)); 
end

indpos = ind(xi>0);
indzero = ind(xi==0);
indpermuted = [indpos; indzero];
[Y,I] = sort(indpermuted); 
U1 = U(indpos, :);
U2 = U(indzero, :); 
disp('det of U2^t*U2 is'); disp(det(U2'*U2));
g2 = U2'\(-U1'*Worstg1);
gperm = [Worstg1;g2]; % g will = gperm(I)
Worstg = gperm(I); 
xipos = xi(xi>0);  % These are the positive design weights
Spos = S(xi>0, :);   % These are the corresponding values of t
bestdesign = [Spos, xipos, Worstg(xi>0)];
%% Display final output

disp('Best design:');
if(example == 1 || example == 2)
    disp([blanks(8), 'design points', blanks(4), 'weights', blanks(5), 'g']);
elseif(example == 3)
	disp([blanks(4), 'points', blanks(4), 'weights', blanks(5), 'g']);
end
disp(bestdesign);

disp('Best run:');
    disp([blanks(4), 'stage', blanks(6), 'maxW', blanks(6), 'pow', blanks(8),...
        'eff', blanks(5), 'accept', blanks(4), 'improve', blanks(3), 'trials/100']);
disp(Bestrun);

bestDesignIndicator = bestDesignSummary - kron(Bestxi, ones(1,num_runs));
frequencyofbest = num_runs - sum(any(bestDesignIndicator));

disp('Frequency of best design:');
disp(frequencyofbest);

disp('Norm of Worstg1:');
disp(norm(Worstg1));
disp('Norm of Worstg:');
disp(norm(Worstg));

if(omega < 1)
    [bestdiv, divgrad, divhess] = div(Worstg(xi>0), link, mu0(xi>0), xi(xi>0));
    disp('Norm of gradient at final g1:');
    disp(norm(bestdiv));    
    e = diag(divhess);
    disp('Range of e-values of Hessian at final g1:');disp([min(e); max(e)]);
end

% Plot the results
annealplots(trialsInBestRun, Bestrun, Worstg, bestdesign, n, example, S);
%% ============================= SUBFUNCTIONS: =============================

%% ============================= subfunction "annealplots" =============================
% These commands can be run on their own, after the main programme is run -
% all of the required inputs will be in memory
function annealplots(trialsInBestRun, Bestrun, Worstg, bestdesign, n, example, S)

% Plot the annealing progess in the best run
figure(1);
subplot(4,1,1);
trials = trialsInBestRun(:,1);
w = trialsInBestRun(:,2);
BestTrialsInRun = trials(w == max(w));

plot(trials, w);
if(length(BestTrialsInRun) == 1) 
    set(gca,'XTick', BestTrialsInRun);
else
    set(gca,'XTick', [min(BestTrialsInRun), max(BestTrialsInRun)]);
end
title({'accepted W vs. trial in best run;'; ... 
    'ticks at first and last appearances of best design'});

subplot(4,1,2);
MaxW = Bestrun(:,2);
bar(MaxW);
title('W of best design in stage')

subplot(4,1,3);
acceptanceRate = Bestrun(:,5);
bar(acceptanceRate);
title('acceptance rate vs. stage')

subplot(4,1,4);
improvementRate = Bestrun(:,6);
bar(improvementRate);
title('improvement rate vs. stage')

% plot the design:
if(example == 1 || example == 2)
figure(2)
subplot(1,2,1);
X = bestdesign(:,1); 
Y = bestdesign(:,2); 
Z = n*bestdesign(:,3); 
stem3(X,Y,Z, 'fill');
xlabel('Treatment');
ylabel('Age');
zlabel('frequency');
set(gca,'xtick',[0 1]);
set(gca,'ytick',30:15:75);
set(gca,'ztick',0:round(max(Z)));
subplot(1,2,2);
X = S(:,1); 
Y = S(:,2);
Z = Worstg;
stem3(X,Y,Z,'.', 'MarkerSize',16);
xlabel('Treatment');
ylabel('Age');
zlabel('Worst g');
set(gca,'xtick',[0 1]);
set(gca,'ytick',30:15:75);
set(gca,'ztick',round(min(Z)-1):round(max(Z)));

elseif(example == 3)
figure(2)
subplot(1,2,1);
X = bestdesign(:,1);  
Y = n*bestdesign(:,2);
stem(X,Y, 'fill');
xlabel('x');
xlim([min(S)-.05 max(S)+.05]);
ylim([0 max(Y)+1]) 
set(gca,'ytick',0:(max(Y)+1));
ylabel('frequency');
subplot(1,2,2);
X = S; 
Y = Worstg;
stem(X,Y,'.', 'MarkerSize',16);
xlabel('x');
xlim([min(S)-.05 max(S)+.05]);
ylim([min(Y)-1 max(Y)+1]) 
ylabel('Worst g');
end
%% ============================= subfunction "annealrun" =============================
% Carries out one run of the annealing algorithm
function [bestxi, worstg1, maxW, pow, eff, trialSummary, stageSummary] = ...
    annealrun(numStages, maxK, maxKK, Lambda, link, maxm, mu0, U, q, n, N, eta, g1reps, temperature, maxL, omega)
% Initialize:
xi = [];  
W = 0;  % Initialize W(xi) so that anything is an improvement
start = 1;
stage = 1;
improvementRate = 1;
stageSummary = []; % temperature, bestW, power, efficiency, 
                   % acceptanceRate, improvementRate and # of trials in stage
maxW = W;
bestxi = [];
trialSummary = []; % records the number of the trial, the current W (i.e. accepted W) and temperature
trialCountInRun = 0;

while (improvementRate >= 1/(maxL + 1)|| stage <= numStages) % this loop over stages is one run
    L = 0;
    num_better = 0; % Compute improvementRate for each stage
    num_worse = 0;
    num_accept = 0;
    trial = 0;    
    while(L < maxL) % this loop over trials is one stage
       % Obtain a new (or first, if start = 1) trial design, together with the minimum power:
       [newxi, newMinPow, g1] = ...
           newdesign(omega, maxK, maxKK, Lambda, link, start, mu0, xi, q, n, N, eta, g1reps);
       % Incorporate the efficiency:
        newEff = efficiency(newxi, U, mu0, link, maxm);
        newW = (1-omega)*newMinPow + omega*newEff;
        
        % Use this design if it is better; otherwise accept with a certain probability:
        deltaW = newW - W;  % accept newxi if this is positive; otherwise accept with accept_prob: 
        accept_prob = min(1,exp(deltaW/temperature)); % = 1 if deltaW > 0
        if(rand < accept_prob) 
            xi = newxi;
            W = newW;
            num_accept = num_accept + 1;
        end
        % Record the best design so far; reset L to 0 if 
        % there was an improvement, otherwise increase L:
        if(newW > maxW)
            maxW = newW;
            pow = newMinPow;
            eff = newEff;
            bestxi = newxi;
            worstg1 = g1;
            num_better = num_better + 1;
            L = 0;
        else
            num_worse = num_worse + 1;
            L = L + 1;
        end
        start = 0;
        trial = num_better + num_worse;
 %       disp('trial'); disp(trial);
        trialCountInRun = trialCountInRun + 1;
        trialSummary = [trialSummary; trialCountInRun, W, temperature];
    end  % End loop over trials within a stage   
    
	xi = bestxi;
    W = maxW;
    
    % Record stage, temperature, bestMinDiv, acceptanceRate, improvementRate and # of trials for this stage:
    acceptanceRate = num_accept/trial;
    improvementRate = num_better/trial;
    disp([blanks(4), 'stage', blanks(6), 'maxW', blanks(6), 'pow', blanks(8),...
        'eff', blanks(5), 'accept', blanks(4), 'improve', blanks(3), 'trials/100']);
    stageSummary = [stageSummary; stage, maxW, pow, eff, acceptanceRate, improvementRate, trial/100];
    disp(stageSummary);
    % reduce the temperature in each stage:
    temperature = .9*temperature;
    stage = stage + 1;
end   % End loop over stages 


%% ============================= subfunction "newdesign" =============================
% Makes a small random change in the current design; returns the new design
% together with the minimum power

function [newxi, newMinPow, g1] = ...
    newdesign(omega, maxK, maxKK, Lambda, link, start, mu0, xi, q, n, N, eta, g1reps)
ind = (1:N)';
if(start == 1)
    xi = zeros(N,1);
    % Randomly generate an initial design of size n; compute xi = N-vector of relative frequencies
    points = [];
    while(length(unique(points)) < q)  % Choose q distinct indices from 1:N
        points = unidrnd(N, q, 1); 
    end
    frequencies = ones(q,1);
    while(sum(frequencies) < n)
        i = unidrnd(q);
        frequencies(i) = frequencies(i) + 1;
    end
    xi(points) = frequencies;
    xi = xi/n;  % This is the starting design
else        
    nxi = round(n*xi);  % Design frequencies
    indpos = ind(nxi>0);
    I = length(indpos);
    choice = ceil(I*rand);
    location1 = indpos(choice); % choose one of the locations with nxi>0
    nxi(location1) = nxi(location1) - 1; % remove mass from this location
    location2 = ceil(N*rand); % choose any of the possible locations
    nxi(location2) = nxi(location2) + 1;  % add mass to this location
    xi = nxi/n;  % This is the new design
end 

%Can input a specific design here; then use maxL = 2, stages >= 2
%N=length(S);
%[(1:N)', S]
%xi = zeros(N,1);
%xi(49) = 14;
%xi(50) = 1;
%xi(80) = 10;
%xi(82) = 15;
%xi = xi/sum(xi);

newxi = xi;
if(omega < 1)
    % Compute the minimizing vector g1, and the minimum KL-divergence:
    indpos = ind(xi>0);
    s = length(indpos);
    outmat = zeros(s+1, g1reps);
    for count = 1:g1reps
        g1start = randn(s,1); % Random starting point;
        g1start = eta*g1start/norm(g1start);
        [g1, fval] = descent(g1start, maxK, maxKK, Lambda, link, mu0(indpos), xi(indpos), eta);
        outmat(:, count) = [g1; fval];
    end
%	outmat
%    keyboard % See if g1reps is too large or too small
    bestfval = min(outmat(s+1,:));
    g1 = outmat((1:s),(outmat(s+1,:) == bestfval));
    newMinPow = normcdf(sqrt(2*n*bestfval) - norminv(.95));    
else
    % Skip the use of fmincon if omega = 1
    g1 = zeros(length(xi(xi>0)),1);
    newMinPow = 0;
end

%% ============================= subfunction "descent" =============================
% descent carries out the steepest descent procedure
function [g1, fval] = descent(g1start, maxK, maxKK, Lambda, link, mu0, xi, eta)
for kk = 1:maxKK    
[divstart, divgrad] = div(g1start, link, mu0, xi);
d = divgrad/norm(divgrad);
lambda = (1 + Lambda*rand)*2*(d'*g1start); % Lambda > 1 guarantees norm(g1) > eta
k=1;
g1 = g1start;
fval = divstart;
while (k <= maxK && (fval >= divstart || norm(g1) < eta))
	g1 = g1start - lambda*d;
    fval = div(g1, link, mu0, xi);
	lambda = lambda/2;
	k = k+1;
end
if (fval >= divstart || norm(g1) < eta || isnan(fval)) 
    g1 = g1start;
    fval = divstart;
end
%out(:,kk) = [kk; divstart; fval];
g1start = g1;
end
%out
%keyboard % See if KK is too large or too small



%% ============================= subfunction "div" =============================
% div is the divergence D(g;xi)
function [div, divgrad, divhess] = div(g, link, mu0, xi)
mug = mu0 + g;
F0 = F(link, mu0);
Fg = F(link, mug);
Fbar0 = 1-F0;
Fbarg = 1-Fg;
KL = Fg.*log(Fg./F0) + Fbarg.*log(Fbarg./Fbar0);
div = KL'*xi;
if nargout > 1
    deriv1 = f(link, mug).*log(Fg.*Fbar0./(F0.*Fbarg));
    divgrad = xi.*deriv1;
    if nargout > 2
        deriv2 = -psi(link, mug).*deriv1 + v(link, mug);
        divhess = diag(deriv2.*xi);
    end
end

%% ============================= subfunction "gdivcon" =============================
function [xineq, xeq, GCineq, GCeq] = gdivcon(g, eta)
xeq = 0;
xineq = eta^2 - g'*g;
if nargout > 2
    GCineq = -2*g;
    GCeq = zeros(length(g),1);
end

%% ============================= subfunction "efficiency" =============================
function efficiency = efficiency(xi, U, mu0, link, maxm)
M = U'*(diag(xi.*v(link, mu0)))*U;  % Information matrix
q = size(M,1);
efficiency = (det(M)^(1/q))/maxm;

%% ============================= subfunction "psi" =============================
% psi is -f'/f, needed in the Hessian of div
function psi = psi(link, t) 
if(link == 'l') 
    psi = tanh(t/2);
elseif(link == 'p')
    psi = t;
end

%% ============================= subfunction "v" =============================
function v = v(link, t)
% v = f^2/(F*Fbar), needed in the information matrix
v = (f(link, t).^2)./(F(link, t).*(1-F(link, t)));

%% ============================= subfunction "F" =============================
function F = F(link, t) 
if(link == 'l') 
    F = 1./(1+exp(-t));
elseif(link == 'p')
    F = normcdf(t);
end    

%% ============================= subfunction "f" =============================
function f = f(link, t) 
if(link == 'l') 
    f = exp(-t)./((1+exp(-t)).^2);
elseif(link == 'p')
    f = normpdf(t);
end



